import pandas as pd
import numpy as np
import scipy.stats as stats

import bluebelt.styles
import bluebelt.core.helpers

import warnings

def line(index, values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.plot(index, values, label=label, **kwargs, **style.default.line)

def dotted_line(index, values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.plot(index, values, label=label, **kwargs, **style.default.dotted_line)

def dashed_line(index, values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.plot(index, values, label=label, **kwargs, **style.default.dashed_line)
    
def fill_between(index, lower, upper, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.fill_between(index, lower, upper, label=label, **kwargs, **style.default.fill_between)

def scatter(index, values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.plot(index, values, label=label, **kwargs, **style.default.scatter)
    
def observations(index, values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.plot(index, values, label=label, **kwargs, **style.default.observations)
    
def out_of_bounds(index, values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.plot(index, values, label=label, **kwargs, **style.default.out_of_bounds)

def outliers(index, values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.plot(index, values, label=label, **kwargs, **style.default.outliers)

def area(index, values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.stackplot(index, values, **kwargs, **style.default.area)
    ax.plot(index, values, label=label, **kwargs, **style.default.line)

def hist(values, ax, label=None, **kwargs):
    style = kwargs.pop('style', bluebelt.styles.paper)
    ax.hist(values, **kwargs, label=label, **style.default.hist) 

# def boxplot(series, ax, **kwargs):

#     # get data
#     if isinstance(series, (pd.Series, pd.DataFrame))and series.isnull().values.any():
#         warnings.warn('the series contains Null values which will be removed', Warning)
#         series = series.dropna()

#     style = kwargs.pop('style', bluebelt.styles.paper)
        
#     boxplot = ax.boxplot(series)

#     for box in boxplot['boxes']:
#         # add style if any is given
#         box.set(**style.graphs.boxplot.boxplot.get('boxes', {}))



import warnings
import datetime
import pandas as pd
import numpy as np

import bluebelt.core.decorators

@bluebelt.core.decorators.class_methods
@pd.api.extensions.register_index_accessor('blue')
class IndexToolkit():
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def alt(self, **kwargs):
        if isinstance(self._obj, pd.MultiIndex):
                
            if 'week' in self._obj.names:
                index = self._complete(iso=True)
                _index = np.array([datetime.datetime.fromisocalendar(*row).replace(tzinfo=datetime.timezone.utc).timestamp() for row in index], dtype='int')
                
            elif 'month' in self._obj.names:
                index = self._complete(iso=False)
                _index = np.array([datetime.datetime(*row).replace(tzinfo=datetime.timezone.utc).timestamp() for row in index], dtype='int')
        
        elif isinstance(self._obj, pd.DatetimeIndex):
            #_index = np.array([row.timestamp() for row in self._obj], dtype='int')
            return self._obj

        else:
            _index = self._obj

        # clean _index
        _index = _index - _index.min()
        _index = _index / np.gcd.reduce(_index)
        
        return _index

    def clean(self, **kwargs):
        # return a clean index for numpy calculations
        if isinstance(self._obj, pd.MultiIndex):
                
            if 'week' in self._obj.names:
                index = self._complete(iso=True)
                _index = np.array([datetime.datetime.fromisocalendar(*row).replace(tzinfo=datetime.timezone.utc).timestamp() for row in index], dtype='int')
                
            elif 'month' in self._obj.names:
                index = self._complete(iso=False)
                _index = np.array([datetime.datetime(*row).replace(tzinfo=datetime.timezone.utc).timestamp() for row in index], dtype='int')
        
        elif isinstance(self._obj, pd.DatetimeIndex):
            _index = np.array([row.timestamp() for row in self._obj], dtype='int')
            
        else:
            _index = self._obj

        # clean _index
        _index = _index - _index.min()
        _index = _index / np.gcd.reduce(_index)
        
        return _index

    def _complete(self, iso=True, levels=None):
        '''
        add missing levels to the (iso)datetimemultiindex
        '''
        _dict = self._obj.to_frame(index=False).to_dict()
        
        # set default dict
        if iso:
            _default_dict ={
                'year': {key: datetime.datetime.now().year for key in range(self._obj.shape[0])},
                'week': {key: 1 for key in range(self._obj.shape[0])},
                'day': {key: 1 for key in range(self._obj.shape[0])},
            }
        else:
            _default_dict = {
                'year': {key: datetime.datetime.now().year for key in range(self._obj.shape[0])},
                'month': {key: 1 for key in range(self._obj.shape[0])},
                'day': {key: 1 for key in range(self._obj.shape[0])},
                'hour': {key: 0 for key in range(self._obj.shape[0])},
                'minute': {key: 0 for key in range(self._obj.shape[0])},
                'second': {key: 0 for key in range(self._obj.shape[0])},
            }

        # add levels
        _levels = levels or _default_dict.keys()
        for level, data in _default_dict.items():
            if level in _levels:
                _dict[level] = _dict.get(level, _default_dict[level])

        return pd.MultiIndex.from_frame(pd.DataFrame.from_dict(_dict)[_default_dict.keys()])

    def to_isodatetimemultiindex(self, **kwargs):
        '''
        change the pandas.DatetimeIndex to a IsoDatetimeMultiIndex
        '''

        if isinstance(self._obj, pd.MultiIndex):
            return self._obj

        elif isinstance(self._obj, pd.DatetimeIndex):
                
            levels = ['year','week','day']
            level = kwargs.get('level', levels[-1])

            # get the levels
            if isinstance(level, str) and level in levels:
                levels = levels[:levels.index(level)+1]
            elif isinstance(level, list):
                _levels = [l for l in level if l in levels]
                if len(_levels) > 0:
                    levels = _levels

            frame = self._obj.isocalendar()[levels]

            return pd.MultiIndex.from_frame(frame[levels])

    iso = to_isodatetimemultiindex

    def to_datetimemultiindex(self, **kwargs):
        '''
        change the pandas.DatetimeIndex to a DatetimeMultiIndex
        '''

        if isinstance(self._obj, pd.MultiIndex):
            return self._obj
            
        elif isinstance(self._obj, pd.DatetimeIndex):
                        
            _dict = {
                'year': self._obj.year, # year including the century
                'month': self._obj.month, # month (1 to 12)
                'day': self._obj.day, # day of the month (1 to 31)
                #'hour': self._obj.hour, # hour, using a 24-hour clock (0 to 23)
                #'minute': self._obj.minute,
                #'second': self._obj.second,
            }
            
            level = kwargs.get('level', list(_dict.keys())[-1])

            # filter _dict with the levels
            if isinstance(level, str) and level in _dict.keys():
                _dict = {key: _dict[key] for key in list(_dict.keys())[:list(_dict.keys()).index(level)+1]}
            elif isinstance(level, list):
                _t_dict = {k: v for k, v in _dict.items() if k in level}
                if len(_t_dict) > 0:
                    _dict = _t_dict

            return pd.MultiIndex.from_frame(pd.DataFrame(_dict))
    
    dt = to_datetimemultiindex


import bluebelt.styles.defaults as defaults

# main plot | pyplot.plot
plot_quantity = {
    'color': defaults.black,
    'linestyle': 'solid',
    'linewidth': 1,
    'zorder': 90,
}
# main plot | pyplot.fill_between
fill_between_quantity = {
    'color': None,
    'hatch': '////',
    'linestyle': 'dashed',
    'linewidth': 0.5,
    'zorder': 50,
}

# main plot | pyplot.plot
plot_distribution = {
    'color': defaults.blue,
    'linestyle': 'solid',
    'linewidth': 1,
    'zorder': 89,
}

# main plot | pyplot.fill_between
fill_between_distribution = {
    'color': None,
    'edgecolor': defaults.blue,
    'hatch': '\\\\\\\\',
    'linestyle': 'dashed',
    'linewidth': 0.5,
    'zorder': 49,
}

plot_skills = {
    'color': defaults.green,
    'linestyle': 'solid',
    'linewidth': 1,
    'zorder': 88,
}
# main plot | pyplot.fill_between
fill_between_skills = {
    'color': None,
    'edgecolor': defaults.green,
    'hatch': '||||',
    'linestyle': 'dashed',
    'linewidth': 0.5,
    'zorder': 48,
}

# main plot | pyplot.plot
plot_qds = {
    'color': defaults.red,
    'linestyle': 'solid',
    'linewidth': 1,
    'zorder': 100,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}


import re
import math
from decimal import Decimal

import numpy as np
import pandas as pd
from scipy.stats import distributions

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from itertools import groupby

# matplotlib format
def _axisformat(ax, series):

    if isinstance(series.index, pd.DatetimeIndex):
        rng = series.index.max() - series.index.min()

        if rng > pd.Timedelta('365 days'):
            ax.xaxis.set_major_locator(mdates.MonthLocator(interval=3))
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%m-%Y'))
            ax.set_xlabel('month')
        elif rng > pd.Timedelta('183 days'):
            ax.xaxis.set_major_locator(mdates.MonthLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%b'))
            ax.set_xlabel('month')
        elif rng > pd.Timedelta('31 days'):
            ax.xaxis.set_major_locator(mdates.WeekdayLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%V'))
            ax.set_xlabel('week')
        else:
            ax.xaxis.set_major_locator(mdates.DayLocator(interval=7))
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%d-%m-%Y'))
            ax.set_xlabel('date')
    elif isinstance(series, pd.DataFrame):
        ax.set_xlabel(series.iloc[:,0].name)
        ax.set_ylabel(series.index.name)
    else:
        ax.set_xlabel(series.name)

    return ax

import matplotlib.transforms as transforms

def _get_multiindex_xticks(index, level, factor):
    
    codes = index.codes[level]
    
    # build values
    values = np.array([index.levels[level][code] for code in codes])
    label_values = values[np.where(np.append(values[:-1] != values[1:], True))[0]]
    values = (np.ceil((values / factor) - 1) * factor) + 1
    
    
    sub_values = np.concatenate(([0,], (np.where(values[:-1] != values[1:])[0]) + 1, [len(values)]))

    # get interval medians
    median_func = lambda x: (x-1)/2
    intervals = sub_values[1:]-sub_values[:-1]
    intervals = median_func(intervals)

    # get new xticks
    xticks = intervals + sub_values[:-1]

    # make new labels
    label_values_grouped = (np.ceil((label_values / factor) - 1) * factor) + 1
    index_min = np.concatenate(([0,], (np.where(label_values_grouped[:-1] != label_values_grouped[1:])[0]) + 1))
    index_max = np.concatenate((np.where(label_values_grouped[:-1] != label_values_grouped[1:])[0], [len(label_values_grouped) - 1]))
    s1 = np.maximum(np.array([label_values[i] for i in index_min]), np.array([label_values_grouped[i] for i in index_min])).astype(int)
    s2 = np.array([label_values[i] for i in index_max]).astype(int)
    if factor > 1:
        # check if the last tick value is in s1 or s2
        if not label_values[-1] in np.concatenate((s1, s2), 0):
            s2 = np.append(s2, label_values[-1])
        xticklabels = [f'{s1[i]}-{s2[i]}' if (s1[i]!=s2[i] and i < len(s2)) else f'{s1[i]}' for i in range(len(s1))]
    else:
        xticklabels = label_values

    # calculate location of lines
    line_locations = np.concatenate(([1,], np.diff(values, n=1)[:-1])).nonzero()[0]
    line_locations = (np.concatenate((line_locations, [s.shape[0]])))
    line_locations = line_locations - 0.5
    
    return xticks, xticklabels, line_locations

def _find_multiindex_levels(index, levels):
    names = index.names
    if isinstance(levels, str):
        levels = names.index(levels)
    elif isinstance(levels, int):
        levels = levels
    elif isinstance(levels, (list, tuple)):
        levels = sorted([_find_index_levels(index=index, levels=level) for level in levels], reverse=True)
        
    if not isinstance(levels, list):
        levels = [levels]
    return levels

def _add_line(ax, xpos, ypos, length, style):
    trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
    line = plt.Line2D([xpos, xpos], [ypos - length, ypos], transform=trans, **style.graphs.line.multiindex_label_line)
    line.set_clip_on(False)
    ax.add_line(line)
    
def set_multiindex_xticks(ax, index, levels=[0, 1], factor=1, **kwargs):
    style = kwargs.pop('style', {'color': 'black'})

    trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
    # get the proper levels
    levels = _find_multiindex_levels(index=index, levels=levels)
    
    # hide the old ticks
    ax.tick_params(axis='x', length=0)
    
    for i, level in enumerate(levels):
        if i == 0:
            xticks, xticklabels, line_locations = _get_multiindex_xticks(index=index, level=level, factor=factor)
            
            # set real xticks
            ax.set_xticks(xticks)
            ax.set_xticklabels(xticklabels)
            
            # draw lines between xticks
            ypos = 0.005
            length = 0.01
            for xpos in line_locations:
                _add_line(ax, xpos, ypos, length, style)
                
        else:
            xticks, xticklabels, line_locations = _get_multiindex_xticks(index=index, level=level, factor=1)
            
            label_pos = -0.08
            # set extra xticks
            for tick, label in zip(xticks, xticklabels):
                ax.text(tick, label_pos, label, ha='center', transform=trans)
            
            # draw lines between xticks
            ypos = 0.005
            length = 0.08
            for xpos in line_locations:
                _add_line(ax, xpos, ypos, length, style)
             
def _get_multiindex_labels(ax, _obj, style):
    # great thanks to Trenton McKinney
    # https://stackoverflow.com/questions/19184484/how-to-add-group-labels-for-bar-charts-in-matplotlib

    def _add_line(ax, xpos, ypos):
        line = plt.Line2D([xpos, xpos], [ypos + .025, ypos], transform=ax.transAxes, **style.graphs.line.multiindex_label_line)
        line.set_clip_on(False)
        ax.add_line(line)

    def _get_label_len(my_index,level):
        labels = my_index.get_level_values(level)
        return [(k, sum(1 for i in g)) for k,g in groupby(labels)]

    # remove current labels
    ax.set_xticklabels([])
    ax.set_xlabel('')

    # draw new labels
    ypos = -.025
    
    # margins
    xmargins = 2 * ax.margins()[0]

    scale = (1.-xmargins)/(_obj.index.size - 1)
    for level in range(_obj.index.nlevels)[::-1]:
        pos = 0
        for label, rpos in _get_label_len(_obj.index,level):
            #lxpos = ((pos + .5 * rpos)*scale) + ax.margins()[0]
            #ax.text(lxpos, ypos, label, ha='center', transform=ax.transAxes)
            _add_line(ax, (pos*scale) + ax.margins()[0], ypos)
            pos += rpos
        #_add_line(ax, (pos*scale) + ax.margins()[0] , ypos)
        ypos -= .1


def _get_multiindex_labels_old(ax, _obj, style):
    # great thanks to Trenton McKinney
    # https://stackoverflow.com/questions/19184484/how-to-add-group-labels-for-bar-charts-in-matplotlib

    def _add_line(ax, xpos, ypos):
        line = plt.Line2D([xpos, xpos], [ypos + .1, ypos], transform=ax.transAxes, **style.graphs.line.multiindex_label_line)
        line.set_clip_on(False)
        ax.add_line(line)

    def _get_label_len(my_index,level):
        labels = my_index.get_level_values(level)
        return [(k, sum(1 for i in g)) for k,g in groupby(labels)]

    # remove current labels
    ax.set_xticklabels([])
    ax.set_xlabel('')

    # draw new labels
    ypos = -.1
    #
    #
    #
    #
    scale = (1.-(2 * ax.margins()[0]))/_obj.index.size
    #scale = 1./_obj.index.size
    for level in range(_obj.index.nlevels)[::-1]:
        #
        #
        #
        pos = 0
        #
        #
        #
        # pos = 0
        for label, rpos in _get_label_len(_obj.index,level):
            lxpos = ((pos + .5 * rpos)*scale) + ax.margins()[0]
            ax.text(lxpos, ypos, label, ha='center', transform=ax.transAxes)
            _add_line(ax, (pos*scale) + ax.margins()[0], ypos)
            pos += rpos
        _add_line(ax, (pos*scale) + ax.margins()[0] , ypos)
        ypos -= .1


def _set_x_axis(ax, _obj):
    # set the x axis, depending on the index data type


    # xticks
    # if isinstance(_obj.series.index, pd.MultiIndex):
    #     bluebelt.graph.helpers._get_multiindex_labels(ax1, _obj.series, style)
    #     fig.subplots_adjust(bottom=.1*_obj.series.index.nlevels)
    # elif isinstance(_obj.series.index, pd.DatetimeIndex):
    #     bluebelt.graph.helpers._axisformat(ax1, _obj.series)
    
    
    #bluebelt.graph.helpers._axisformat(ax1, _obj.series)


    return 


import pandas as pd
import numpy as np
import scipy.stats as stats

import warnings

import matplotlib.pyplot as plt

import bluebelt.helpers.xticks
import bluebelt.core.index
import bluebelt.core.decorators
import bluebelt.graph.helpers

import bluebelt.data.resolution

import bluebelt.styles

@bluebelt.core.decorators.class_methods


class Effort():
    """
    Calculate the planning effort
    """
    def __init__(self, series, level=['year', 'week'], **kwargs):
        
        self.series = series
        self.level = level

        self.calculate()

    def calculate(self):

        self.quantity = bluebelt.data.resolution.GroupByDatetimeIndex(self.series, level=self.level, complete=True).diff_quantity()
        self.distribution = bluebelt.data.resolution.GroupByDatetimeIndex(self.series, level=self.level, complete=True).diff_distribution()
        if isinstance(self.series, pd.DataFrame):
            self.skills = bluebelt.data.resolution.GroupByDatetimeIndex(self.series, level=self.level, complete=True).diff_skills()
        else:
            self.skills = pd.Series(index=self.quantity.index, data=np.zeros(self.quantity.index.shape[0]), name='skills')

        self.qds = 1 - ((1 - self.quantity) * (1 - self.distribution) * (1 - self.skills))
    def __repr__(self):
        return (f'{self.__class__.__name__}(n={self.series.shape[0]:1.0f}, qds={self.qds.mean():1.4f}, quantity={self.quantity.mean():1.4f}, distribution={self.distribution.mean():1.4f}, skills={self.skills.mean():1.4f})')
    
    def plot(self, **kwargs):
        
        return _qds_plot(self, **kwargs)

def _qds_plot(_obj, **kwargs):
        
    style = kwargs.pop('style', bluebelt.styles.paper)
    title = kwargs.pop('title', f'planning effort QDS plot')
    
    group = kwargs.pop('group', None)
    path = kwargs.pop('path', None)
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (0, 1))
    
    # prepare figure
    fig, axes = plt.subplots(nrows=1, ncols=1, **kwargs)

    _index = bluebelt.core.index.IndexToolkit(_obj.quantity.index[1:]).alt()

    # q
    axes.fill_between(_index, 0, _obj.quantity.values[1:], **style.planning.fill_between_quantity, label='quantity')
    axes.plot(_index, _obj.quantity.values[1:], **style.planning.plot_quantity)
    
    # d
    axes.fill_between(_index, 0, _obj.distribution.values[1:], **style.planning.fill_between_distribution, label='distribution')
    axes.plot(_index, _obj.distribution.values[1:], **style.planning.plot_distribution)

    # s
    axes.fill_between(_index, 0, _obj.skills.values[1:], **style.planning.fill_between_skills, label='skills')
    axes.plot(_index, _obj.skills.values[1:], **style.planning.plot_skills)
    
    # qds
    axes.plot(_index, _obj.qds.values[1:], **style.planning.plot_qds, label='qds')
    
    # format things
    axes.set_ylim(ylim)

    # set xticks
    bluebelt.helpers.xticks.set_xticks(ax=axes, index=_obj.quantity.index[1:], location=_index[1:], group=group)

    # transform yticklabels to percentage
    axes.set_yticks(axes.get_yticks())
    axes.set_yticklabels([f'{y:1.0%}' for y in axes.get_yticks()])

    # title
    axes.set_title(title, **style.graphs.line.title)

    # legend
    axes.legend(loc='upper left')

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig


import bluebelt.styles.defaults as defaults


# pattern plot | pyplot.plot
line = {
    'marker': None,
    'linewidth': 1,
    'color': defaults.blue+(1,),
    'linestyle': 'solid',
}

dotted_line = {
    'marker': None,
    'linewidth': 1,
    'color': defaults.blue,
    'linestyle': 'dotted',
    'linewidth': 1,
}

dashed_line = {
    'marker': None,
    'linewidth': 1,
    'color': defaults.blue,
    'linestyle': 'dashed',
    'linewidth': 1,
}
scatter = {
    'linewidth': 0,
    'marker': 'o',
    'markersize': 7,
    'markerfacecolor': defaults.blue,
    'markeredgewidth': 0,
    'markeredgecolor': defaults.blue,
    'alpha': 0.7,
}

observations = {
    'linewidth': 0,
    'marker': 'o',
    'markersize': 5,
    'markerfacecolor': defaults.black,
    'markeredgewidth': 0,
    'markeredgecolor': defaults.black,
    'alpha': 1,
}

out_of_bounds = {
    'linewidth': 0,
    'marker': 'o',
    'markersize': 5,
    'markerfacecolor': defaults.blue,
    'markeredgewidth': 0,
    'markeredgecolor': defaults.blue,
    'alpha': 1,
}

outliers = {
    'linewidth': 0,
    'marker': '+',
    'markersize': 7,
    'markerfacecolor': defaults.red,
    'markeredgewidth': 1,
    'markeredgecolor': defaults.red,
    'alpha': 1,
}

area = {
    'color': None,
    'hatch': '////',
    'linestyle': 'dashed',
    'linewidth': 0.5,
}

fill_between = {
    'color': None,
    'hatch': '////',
    'edgecolor': defaults.blue+(1,),
    'linestyle': '-',
    'linewidth': 0,
}

hist = {
    'density': True, 
    'alpha': 1,
    'edgecolor': None,
    'facecolor': defaults.blue,
    'fill': False,
    'hatch': '//',
    'linestyle': 'solid',
    'linewidth': 1,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}


import pandas as pd
import numpy as np
import math

import matplotlib.pyplot as plt

from bluebelt.core.checks import check_kwargs

import bluebelt.core.decorators

import bluebelt.core.index
import bluebelt.core.series
import bluebelt.core.dataframe

import warnings

def resolution_methods(cls):

    def sum(self):
        result = self.grouped.sum(min_count=1)
        return result

    def mean(self):
        result = self.grouped.mean()
        return result

    def var(self):
        result = self.grouped.var()
        return result
            
    def std(self):
        result = self.grouped.std()
        return result

    def min(self):
        result = self.grouped.min()
        return result

    def max(self):
        result = self.grouped.max()
        return result

    
    def count(self):
        result = self.grouped.count()
        return result
    
    def value_range(self):
        result = self.grouped.apply(lambda x: x.max() - x.min())
        return result

    def diff_quantity(self):
        # if this week > last week: this week - last week / this week
        # if this week < last week: last week - this week / last week

        if isinstance(self._obj, pd.Series):
            result = (self.grouped.sum() - self.grouped.sum().shift()).abs() / np.maximum(self.grouped.sum(), self.grouped.sum().shift().fillna(0))
        else:
            result = (self.grouped.sum().sum(axis=1) - self.grouped.sum().sum(axis=1).shift()).abs() / np.maximum(self.grouped.sum().sum(axis=1), self.grouped.sum().sum(axis=1).shift().fillna(0))

        return result

    def diff_distribution(self, **kwargs):

        index_groups = self.grouped.obj.index.names
        _level = self.level[-1] if isinstance(self.level, list) else self.level
        group = index_groups[self.grouped.obj.index.names.index(_level) + 1]
            
        if self.grouped.obj.index.names[-1] != _level:
            if isinstance(self._obj, pd.Series):
                result = self.grouped.apply(lambda s: s.groupby(group).sum()).unstack(level=-1)
                result = pd.Series((result - result.shift().multiply((result.sum(axis=1) / result.sum(axis=1).shift()), axis=0)).abs().sum(axis=1, min_count=1) / (result.sum(axis=1) * 2), name='distribution')    
                
            else:
                result = self.grouped.apply(lambda s: s.groupby(group).sum()).sum(axis=1).unstack(level=-1)
                result = pd.Series((result - result.shift().multiply((result.sum(axis=1) / result.sum(axis=1).shift()), axis=0)).abs().sum(axis=1, min_count=1) / (result.sum(axis=1) * 2), name='distribution')
        else:
            result = pd.Series(index=self.grouped.sum().index, data=[0]*self.grouped.sum().size, name=self._obj.name)
        
        return result

    def diff_skills(self, **kwargs):
    
        index_groups = self.grouped.obj.index.names
        _level = self.level[-1] if isinstance(self.level, list) else self.level
        group = index_groups[self.grouped.obj.index.names.index(_level) + 1]
            
        if self.grouped.obj.index.names[-1] != _level and isinstance(self._obj, pd.DataFrame):
            this_week = self._obj.groupby(group).shift(0)
            last_week = self._obj.groupby(group).shift(1)

            _ratio = last_week.sum(axis=1).divide(this_week.sum(axis=1))

            # copy _ratio series to ratio colums (mirror this week)
            ratio = pd.DataFrame(index=_ratio.index, data=[])
            for column in this_week.columns:
                ratio[column] = pd.Series(_ratio, name=column)

            # normalise this week
            this_week_normalised = ratio.multiply(this_week)
            absolute_diff = (this_week_normalised - last_week).abs().sum(axis=1)
            absolute_diff = absolute_diff.groupby(['year', 'week']).sum()
            this_week_normalised = this_week_normalised.sum(axis=1).groupby(['year', 'week']).sum()

            # divide by 2 because 100% shift of skills would otherwise result in 200%; 100% from skill x + 100% to skill y
            result = (absolute_diff / (this_week_normalised * 2))
            
        else:
            result = pd.Series(index=self.grouped.sum().index, data=[0]*self.grouped.sum().size, name=self._obj.name)
        
        return result

    setattr(cls, 'sum', sum)
    setattr(cls, 'mean', mean)
    setattr(cls, 'var', var)
    setattr(cls, 'std', std)
    setattr(cls, 'min', min)
    setattr(cls, 'max', max)
    setattr(cls, 'count', count)
    setattr(cls, 'value_range', value_range)
    setattr(cls, 'diff_quantity', diff_quantity)
    setattr(cls, 'diff_distribution', diff_distribution)
    setattr(cls, 'diff_skills', diff_skills)
    return cls

@bluebelt.core.decorators.class_methods
@resolution_methods
class GroupByDatetimeIndex():
    """
    Group a pandas.Series or pandas.DataFrame by DateTime index and apply a specific function.
        arguments
        series: pandas.Series
        how: str
            a string with date-time keywords that can be parsed to group the index
            keywords:
            
            default value "week"

        Apply one of the following functions:
            .sum()
            .mean()
            .min()
            .max()
            .std()
            .value_range()
            .count()
            .subsize_count()

        e.g. series.blue.data.group_index(how="week").sum()
        
    """
    
    def __init__(self, _obj, level="week", complete=False, **kwargs):

        self._obj = _obj
        self.level = level
        self.complete = complete
        self.nrows = self._obj.shape[0]
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        
        # convert to pd.MultiIndex
        if isinstance(self._obj, pd.Series):
            if 'month' in self.level:
                self._obj = bluebelt.core.series.SeriesToolkit(self._obj).index.to_datetimemultiindex()
            else:
                self._obj = bluebelt.core.series.SeriesToolkit(self._obj).index.to_isodatetimemultiindex()
        elif isinstance(self._obj, pd.DataFrame):
            if 'month' in self.level:
                self._obj = bluebelt.core.dataframe.DataFrameToolkit(self._obj).index.to_datetimemultiindex()
            else:
                self._obj = bluebelt.core.dataframe.DataFrameToolkit(self._obj).index.to_isodatetimemultiindex()

        # check the index type
        if not isinstance(self._obj.index, pd.MultiIndex):
            # convert to isodatetimemultiindex or datetimemultiindex
            if self.iso:
                index = bluebelt.core.index.IndexToolkit(self._obj.index, **kwargs).iso()
            else:
                index = bluebelt.core.index.IndexToolkit(self._obj.index, **kwargs).dt()
        else:
            index = self._obj.index

        self.grouped = self._obj.groupby(level=self.level)

    def __str__(self):
        return ""
    
    def __repr__(self):
        return self.grouped.__repr__()


    def subsize_count(self, count=3, size=1):

            """
            Count the number of times a list of <count> items with size <size> fit in the groupby object (which is a pandas Series)
            e.g.
            groupby object: pd.Series([10, 8, 3, 3, 5])
            count = 3
            size = 1

            returns 9

            step 0: (3, 3, 5, 8, 10)
            step 1: (3, 3, 4, 7, 9)
            step 2: (3, 3, 3, 6, 8)
            step 3: (2, 3, 3, 5, 7)
            step 4: (2, 2, 3, 4, 6)
            step 5: (2, 2, 2, 3, 5)
            step 6: (1, 2, 2, 2, 4)
            step 7: (1, 1, 1, 2, 3)
            step 8: (0, 1, 1, 1, 2)
            step 9: (0, 0, 0, 1, 1)

            """
            if isinstance(count, (float, int)):
                count = [int(count)]
            
            result = {}
            for c in count:
                result[c] = self.grouped.apply(lambda x: _subsize_count(series=x, count=c, size=size)).values
            result = pd.DataFrame(result, index=self.grouped.groups.keys()) #self.grouped.apply(lambda x: _subsize_count(series=x, count=count, size=size))
            
            if len(count) == 2:
                _dict = {}
                for val in range(int(result.values.min()), int(result.values.max())):

                    under = np.where(val < result.iloc[:,0:2].min(axis=1), result.iloc[:,0:2].min(axis=1) - val, 0)
                    over = np.where(val > result.iloc[:,0:2].max(axis=1), val - result.iloc[:,0:2].max(axis=1), 0)
                    _dict[val] = under + over

                # get the keys with the most zeros
                _o_dict = {key: len([x for x in value if x == 0]) for key, value in _dict.items()}
                most_zeros = _o_dict.get(max(_o_dict, key=_o_dict.get))

                _dict = {key: value for (key, value) in _dict.items() if _o_dict.get(key) == most_zeros}

                # get the smallest sum
                _o_dict = {key: sum(value) for (key, value) in _dict.items()}
                min_val = _o_dict.get(min(_o_dict, key=_o_dict.get))
                _dict = {key: value for (key, value) in _dict.items() if _o_dict.get(key) == min_val}

                # build the optimum series and merge with result
                optimum = tuple(_dict.keys()) if len(_dict.keys()) > 1 else list(_dict.keys())[0]
                optimum = pd.Series(index=result.index, data=[optimum]*result.shape[0], name='optimum')

                result = result.merge(optimum, left_index=True, right_index=True)

            return result

def _subsize_count(series, count=3, size=1):
    series = pd.Series(series)/size
    result = series.sum()*count
    for i in range(count, 0, -1):
        result = min(result, math.floor(series.nsmallest(len(series) - count + i).sum() / i))
    return result

def _subseries_count(series, subseries=None, **kwargs):
    series = pd.Series(series)
    subseries = pd.Series(subseries)
    result=series.sum()*subseries.sum()
    for i in range(len(subseries), 0, -1):
        result = min(result, math.floor(series.nsmallest(len(series) - len(subseries) + i).sum() / subseries.nsmallest(i).sum()))
    return result


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import matplotlib.dates as mdates

import bluebelt.styles.defaults

def set_xticks(ax=None, index=None, location=None, group=None):
    if isinstance(index, pd.MultiIndex):
        set_multiindex_xticks(ax=ax, index=index, group=group)
    elif isinstance(index, pd.DatetimeIndex):
        set_datetimeindex_xticks(ax=ax, index=index)
    else:
        ax.set_xticks(location)
        ax.set_xticklabels(index)

# datetimeindex
def set_datetimeindex_xticks(ax, index):
    '''
    set the xticks for a pandas DatetimeIndex
    '''

    rng = index.max() - index.min()
    if rng > pd.Timedelta('365 days'):
        ax.xaxis.set_major_locator(mdates.MonthLocator(interval=3))
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%m-%Y'))
        ax.set_xlabel('month')
    elif rng > pd.Timedelta('183 days'):
        ax.xaxis.set_major_locator(mdates.MonthLocator())
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%b'))
        ax.set_xlabel('month')
    elif rng > pd.Timedelta('31 days'):
        ax.xaxis.set_major_locator(mdates.WeekdayLocator())
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%V'))
        ax.set_xlabel('week')
    else:
        ax.xaxis.set_major_locator(mdates.DayLocator(interval=7))
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%d-%m-%Y'))
        ax.set_xlabel('date')

# multiindex
def set_multiindex_xticks(ax, index, levels=[0, 1], group=None):
    trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)

    # get the proper levels
    levels = _find_multiindex_levels(index=index, levels=levels)
        
    # hide the old ticks
    ax.tick_params(axis='x', length=0)

    # get ax size
    ax_size = ax.get_window_extent().transformed(ax.get_figure().dpi_scale_trans.inverted())
    
    for i, level in enumerate(levels):
        if i == 0:
            
            # calculate group size
            if not group:
                _group = (index.size/(ax_size.width*bluebelt.styles.defaults.small)) * 5.5
                if _group > 1:
                    _group *= 2
                _groups = np.array([1, 4, 13])
                group = _groups[np.greater_equal(_groups - _group, 0).argmax()]

            xticks, xticklabels, line_locations = _get_multiindex_xticks(index=index, level=level, group=group)
            
            # set real xticks
            ax.set_xticks(xticks)
            ax.set_xticklabels(xticklabels)
            
            # draw lines between xticks
            ypos = 0.005
            length = 0.01
            for xpos in line_locations:
                _add_line(ax, xpos, ypos, length)
                
        else:
            xticks, xticklabels, line_locations = _get_multiindex_xticks(index=index, level=level, group=1)

            # get label position compared to ax height
            label_pos = -bluebelt.styles.defaults.small/(ax_size.height*35)

            # set extra xticks
            for tick, label in zip(xticks, xticklabels):
                ax.text(tick, label_pos, label, ha='center', transform=trans)
            
            # draw lines between xticks
            ypos = 0.005
            length = -label_pos
            for xpos in line_locations:
                _add_line(ax, xpos, ypos, length)

    
def _get_multiindex_xticks(index, level, group):
    
    codes = index.codes[level]
    
    # build values
    values = np.array([index.levels[level][code] for code in codes])
    label_values = values[np.where(np.append(values[:-1] != values[1:], True))[0]]
    values = (np.ceil((values / group) - 1) * group) + 1
    
    
    sub_values = np.concatenate(([0,], (np.where(values[:-1] != values[1:])[0]) + 1, [len(values)]))

    # get interval medians
    median_func = lambda x: (x-1)/2
    intervals = sub_values[1:]-sub_values[:-1]
    intervals = median_func(intervals)

    # get new xticks
    xticks = intervals + sub_values[:-1]

    # make new labels
    label_values_grouped = (np.ceil((label_values / group) - 1) * group) + 1
    index_min = np.concatenate(([0,], (np.where(label_values_grouped[:-1] != label_values_grouped[1:])[0]) + 1))
    index_max = np.concatenate((np.where(label_values_grouped[:-1] != label_values_grouped[1:])[0], [len(label_values_grouped) - 1]))
    s1 = np.maximum(np.array([label_values[i] for i in index_min]), np.array([label_values_grouped[i] for i in index_min])).astype(int)
    s2 = np.array([label_values[i] for i in index_max]).astype(int)
    if group > 1:
        # check if the last tick value is in s1 or s2
        if not label_values[-1] in np.concatenate((s1, s2), 0):
            s2 = np.append(s2, label_values[-1])
        xticklabels = [f'{s1[i]}-{s2[i]}' if (s1[i]!=s2[i] and i < len(s2)) else f'{s1[i]}' for i in range(len(s1))]
    else:
        xticklabels = label_values

    # calculate location of lines
    line_locations = np.concatenate(([1,], np.diff(values, n=1)[:-1])).nonzero()[0]
    line_locations = (np.concatenate((line_locations, [index.shape[0]])))
    line_locations = line_locations - 0.5
    
    return xticks, xticklabels, line_locations

def _find_multiindex_levels(index, levels, final_assembly=True):
    names = index.names
    if isinstance(levels, str):
        levels = names.index(levels)
    elif isinstance(levels, int):
        levels = levels
    elif isinstance(levels, (list, tuple)):
        levels = sorted([_find_multiindex_levels(index=index, levels=level, final_assembly=False) for level in levels], reverse=True)
    
    # make the levels a list if 
    if final_assembly and not isinstance(levels, list):
        levels = [levels]
    return levels

def _add_line(ax, xpos, ypos, length):
    trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
    line = plt.Line2D([xpos, xpos], [ypos - length, ypos], color='black', transform=trans)
    line.set_clip_on(False)
    ax.add_line(line)

             


import datetime
import warnings

import pandas as pd

import numpy as np
import scipy.stats as stats

import matplotlib.pyplot as plt

import bluebelt.core.helpers
import bluebelt.core.decorators
import bluebelt.graph.defaults

import bluebelt.styles

import bluebelt.core.index

@bluebelt.core.decorators.class_methods
class Polynomial():
    """
    Find the polynomial of a series and project a bandwidth

    Polynomial(series, shape=(0, 6), validation='p_value', threshold=0.05, confidence=0.8, outlier_sigma=2, adjust=True, **kwargs)
    
    series: pandas.Series
    shape: int or tuple
        when an int is provided the polynomial is provided as n-th degree polynomial
        when a tuple is provided the function will find an optimised polynomial between first and second value of the tuple
        default value: (0, 6)
    validation: string
        validation type for shape tuple
        p_val: test for normal distribution of the residuals
        rsq: check for improvement of the rsq value
        default value: p_val
    threshold: float
        the threshold for normal distribution test or rsq improvement
        default value: 0.05
    confidence: float
        the bound confidence
        default value: 0.8
    outlier_sigma: float
        outliers are datapoints outside the outlier_sigma fraction
        default value: 2
    adjust: boolean
        adjust polynomial for outliers
        default value: True

    """
    def __init__(self, series, shape=(0, 6), validation='rsq', threshold=0.05, confidence=0.8, outlier_sigma=2, adjust=True, **kwargs):
        
        self.series = series
        self.name = series.name or 'series'
        self.shape = shape
        self.validation = validation
        self.threshold = threshold
        self.confidence = confidence
        self.outlier_sigma = outlier_sigma
        self.adjust = adjust

        self.calculate()

    def calculate(self):
        
        # set pattern and residuals
        _poly_hand_granade(self)

        # set outliers
        self.outliers = pd.Series(data=np.where(self.residuals.abs() > self.residuals.std() * self.outlier_sigma, self.series, None), index=self.series.index)
        self.outliers_count = np.count_nonzero(self.outliers)

        # handle adjusted
        self.adjusted = self.series.loc[~self.outliers.notnull()]
        if self.adjust:
            
            self._series = self.series.copy() # backup
            
            # replace outliers with None values so they will be ignored by _poly_hand_granade and reset pattern
            self.series = pd.Series(data=np.where(self.outliers.notnull(), None, self.series).astype(np.float), index=self.series.index)
            _poly_hand_granade(self)

            self.series = self._series.copy() # and reset
            del self._series

        # handle bounds
        _calculate_bounds(self)

        # set final shape
        if hasattr(self, '_shape'):
            self.shape = self._shape
            del self._shape


    def __repr__(self):
        return (f'{self.__class__.__name__}(n={self.series.size:1.0f}, shape={self.shape}, validation=\'{self.validation}\', threshold={self.threshold}, confidence={self.confidence}, outlier_sigma={self.outlier_sigma}, adjust={self.adjust}, outliers={self.outliers_count}, rsq={self.rsq:1.2f}, std={self.std:1.2f}, p_value={self.p_value:1.2f})')
    
    def __str__(self):
        _result = f'input variables\n'
        _result += f'-'*50 + '\n'
        _result += f'  {"series size:":<30}{self.series.size:1.0f}\n'
        _result += f'  {"validation type:":<30}{self.validation}\n'
        _result += f'  {"validation threshold:":<30}{self.threshold:1.4f}\n'
        
        _result += f'\n'
        _result += f'pattern\n'
        _result += f'-'*50 + '\n'
        _result += f'  {"shape:":<30}{self.shape:1.0f}\n'
        _result += f'  {"r squared:":<30}{self.rsq:1.2f}\n'
        

        _result += f'\n'
        _result += f'residuals\n'
        _result += f'-'*50 + '\n'
        _result += f'  {"bounds level:":<30}{self.confidence * 100:1.0f}%\n'
        _result += f'  {"bounds size:":<30}{self.bounds * 2:1.2f}\n'
        _result += f'  {"standard deviation:":<30}{self.std:1.2f}\n'
        _result += f'  {"p-value normal distribution:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"outliers:":<30}{self.outliers_count:1.0f}\n'
        
        return _result

    def plot(self, **kwargs):
        
        return _pattern_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class Periodical():
    
    """
    Find the periodical pattern of a series and project a bandwidth

    Periodical(series, rule='1W', how='mean', resolution=None, confidence=0.8, outlier_sigma=2, adjust=True, **kwargs)
    
    series: pandas.Series
    rule: period representation used for resampling the series
        default value: "1W"
    how: define how the period must be evaluated
        options are "mean", "min", "max" and "std"
        default value: "mean"
    resolution: define the resolution of the pattern
        the pattern is rounded to fit the resolution
        default value: None
    confidence: float
        the bandwidth confidence
        default value: 0.8
    outlier_sigma: float
        outliers are datapoints outside the outlier_sigma fraction
        default value: 2
    
    """

    def __init__(self, series, rule='1W', how='mean', resolution=None, confidence=0.8, outlier_sigma=2, adjust=True, **kwargs):
        
        self.series = series
        self.rule = rule
        self.how = how
        self.resolution = resolution
        self.confidence = confidence
        self.outlier_sigma = outlier_sigma
        self.adjust = adjust
        
        self.calculate()

    def calculate(self):
        
        # set pattern and residuals        
        self.result, self.residuals, self.statistic, self.p_value, self.rsq = _peri_hand_granade(series=self.series, rule=self.rule, how=self.how, resolution=self.resolution)

        # set outliers
        self.outliers = pd.Series(data=np.where(self.residuals.abs() > self.residuals.std() * self.outlier_sigma, self.series, None), index=self.series.index, name=f'{self.series.name} {_get_nice_polynomial_name(self._shape)} outliers')
        self.outliers_count = np.count_nonzero(self.outliers)

        # handle adjusted
        self.adjusted = self.series.loc[~self.outliers.notnull()]
        if self.adjust:
            # replace outliers with None values so they will be ignored by _peri_hand_granade and reset pattern
            values = pd.Series(data=np.where(self.outliers.notnull(), None, self.series).astype(np.float), index=self.series.index)
            self.result, self.residuals, self.statistic, self.p_value, self.rsq = _peri_hand_granade(series=values, rule=self.rule, how=self.how, resolution=self.resolution)
        
        self.result.rename(f'periodical ({self.rule})', inplace=True)

        # handle bounds
        _calculate_bounds(self)

        # 

    def __repr__(self):
        return (f'{self.__class__.__name__}(n={self.series.size:1.0f}, rule={self.rule}, how={self.how}, resolution={self.resolution}, confidence={self.confidence}, outlier_sigma={self.outlier_sigma}, adjust={self.adjust}, outliers={self.outliers_count})')
    
    def plot(self, **kwargs):
        return _pattern_plot(self, **kwargs)


# helper functions
def _poly_hand_granade(_obj):

    # get index and _index
    index = bluebelt.core.index.IndexToolkit(_obj.series.index).clean()
    _index = bluebelt.core.index.IndexToolkit(_obj.series.dropna().index).clean()
    
    # get the values
    _values = _obj.series.dropna().values

    if isinstance(_obj.shape, int):
        polyfit = np.polynomial.polynomial.polyfit(_index, _values, _obj.shape)
        _obj.result = pd.Series(index=_obj.series.index, data=np.polynomial.polynomial.polyval(index, polyfit), name=f'{_obj.name} {_get_nice_polynomial_name(_obj.shape)}')
        _obj.residuals = (_obj.series - _obj.result).rename(f'{_obj.name} {_get_nice_polynomial_name(_obj.shape)} residuals')
                    
        _obj.statistic, _obj.p_value = stats.normaltest(_obj.residuals.dropna().values)
        np_err = np.seterr(divide='ignore', invalid='ignore') # ignore possible divide by zero
        _obj.rsq = np.corrcoef(_obj.series.dropna().values, _obj.result[~_obj.series.isna()].values)[1,0]**2
        np.seterr(**np_err) # go back to previous settings
        _obj.std = _obj.residuals.std()
        
    elif isinstance(_obj.shape, tuple):
        
        _results = {}
        _validation = {}

        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            for shape in range(_obj.shape[0], _obj.shape[1]+1):
                try:
                    polyfit = np.polynomial.polynomial.polyfit(_index, _values, shape)
                    result = pd.Series(index=_obj.series.index, data=np.polynomial.polynomial.polyval(index, polyfit), name=f'{_obj.name} {_get_nice_polynomial_name(shape)}')
                    residuals = (_obj.series - result).rename(f'{_obj.name} {_get_nice_polynomial_name(shape)} residuals')
                    
                    statistic, p_value = stats.normaltest(residuals.dropna().values)
                    np_err = np.seterr(divide='ignore', invalid='ignore') # ignore possible divide by zero
                    rsq = np.corrcoef(_obj.series.dropna().values, result[~_obj.series.isna()].values)[1,0]**2
                    np.seterr(**np_err) # go back to previous settings
                    std = residuals.std()

                    _results[shape] = {
                        'result': result,
                        'residuals': residuals,
                        'statistic': statistic,
                    }
                    _validation[shape] = {
                        'p_value': p_value,
                        'rsq': rsq,
                        'std': std,
                    }
                    

                except np.polynomial.polynomial.pu.RankWarning:
                    print(f'RankWarning at {_get_nice_polynomial_name(shape)}')
                    break
        
        if _obj.validation == 'p_value':
            validation = pd.DataFrame.from_dict(_validation).loc[_obj.validation]
            
            # if any p_value >= threshold then we have a winner
            if (validation >= _obj.threshold).any():
                shape = validation.idxmax()
            else:
                shape = 0

        elif _obj.validation == 'std':

            # we want a small std of the residuals
            validation = pd.DataFrame.from_dict(_validation).loc[_obj.validation]

            # are there any relevant improvements?
            if (validation.diff().abs()/validation.shift() >= _obj.threshold).any():
                # find the relevant improvements
                relevant = validation.iloc[validation.diff().idxmin():]
                
                # does it get any better than this?
                improvements = (relevant.diff()/relevant.shift(1)).abs() >= _obj.threshold
                improvements.at[validation.diff().idxmin()] = True # because it must be
                
                # where does it stop getting better?
                first_fail = improvements.where(improvements == False).first_valid_index() or improvements.index.max()+1
                
                # so what is the shape?
                shape = improvements[improvements.index<first_fail].index.max()
            else:
                shape = validation.index[0]

        else: # the default
            _obj.validation = 'rsq'
            
            # we want a big rsq value
            validation = pd.DataFrame.from_dict(_validation).loc[_obj.validation]

            # are there any relevant improvements?
            if (validation.diff().abs()/validation.shift() >= _obj.threshold).any():
                # find the relevant improvements
                relevant = validation.iloc[validation.diff().idxmax():]
                
                # does it get any better than this?
                improvements = (relevant.diff()/relevant.shift(1)).abs() >= _obj.threshold
                improvements.at[validation.diff().idxmax()] = True # because it must be
                
                # where does it stop getting better?
                first_fail = improvements.where(improvements == False).first_valid_index() or improvements.index.max()+1
                
                # so what is the shape?
                shape = improvements[improvements.index<first_fail].index.max()
            else:
                shape = validation.index[0]


        _obj.result = _results.get(shape).get('result')
        _obj.residuals = _results.get(shape).get('residuals')
        _obj._shape = shape
        _obj.statistic = _results.get(shape).get('statistic')
        _obj.p_value = _validation.get(shape).get('p_value')
        _obj.rsq = _validation.get(shape).get('rsq')
        _obj.std = _validation.get(shape).get('std')
        
    else:
        _obj.result = None
        _obj.residuals = None

    return

def _peri_hand_granade(series, rule, how, resolution, **kwargs):

    # set pattern and residuals
    if how=='mean':
        pattern = series.resample(rule=rule).mean()
    elif how=='min':
        pattern = series.resample(rule=rule).min()
    elif how=='max':
        pattern = series.resample(rule=rule).max()
    elif how=='std':
        pattern = series.resample(rule=rule).std()
    else:
        pattern = series.resample(rule=rule).sum()
    
    # reindex pattern
    if any([period for period in ['M', 'A', 'Q', 'BM', 'BA', 'BQ', 'W'] if (period == "".join(char for char in rule if not char.isnumeric()))]):
        pattern = pattern.reindex_like(series, method = 'bfill')
    else:
        pattern = pattern.reindex_like(series, method = 'ffill')

    if resolution:
        # adjust for resolution
        pattern = pattern.divide(resolution).round(0).multiply(resolution)
    
    residuals = series - pattern

    statistic, p_value = stats.normaltest(residuals.dropna().values)
    rsq = np.corrcoef(series.values, pattern.values)[1,0]**2

    return pattern, residuals, statistic, p_value, rsq

def _get_nice_polynomial_name(shape):
    if shape==0:
        return 'linear'
    if shape==1:
        return str(shape)+'st degree polynomial'
    elif shape==2:
        return str(shape)+'nd degree polynomial'
    elif shape==3:
        return str(shape)+'rd degree polynomial'
    else:
        return str(shape)+'th degree polynomial'

def _calculate_bounds(_obj):
        
    _obj.sigma_level = stats.norm.ppf(1-(1-_obj.confidence)/2)

    # set bounds
    _obj.upper = _obj.result + _obj.residuals.std() * _obj.sigma_level
    _obj.lower = _obj.result - _obj.residuals.std() * _obj.sigma_level
    _obj.bounds = _obj.residuals.std() * _obj.sigma_level

    # set out of bounds values
    _obj.out_of_bounds = _obj.series[((_obj.series > _obj.upper) | (_obj.series < _obj.lower)) & (_obj.outliers.isnull())]
    _obj.within_bounds = _obj.series[(_obj.series <= _obj.upper) & (_obj.series >= _obj.lower)]
    

    return _obj

def _pattern_plot(_obj, **kwargs):
        
    style = kwargs.pop('style', bluebelt.styles.paper)
    path = kwargs.pop('path', None)

    bounds = kwargs.pop('bounds', True)
    residuals = kwargs.pop('residuals', False)
    legend = kwargs.pop('legend', True)
    
    group = kwargs.pop('group', None)
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (None, None))
    
    # prepare figure
    fig = plt.figure(constrained_layout=False, **kwargs)
    if residuals:
        gridspec = fig.add_gridspec(nrows=2, ncols=1, height_ratios=[5,3], wspace=0, hspace=0)
        ax2 = fig.add_subplot(gridspec[1, 0], zorder=40)

        # residuals histogram
        bluebelt.graph.defaults.hist(_obj.residuals.values, ax=ax2)
        ax2.set_yticks([])

        # get current limits
        xlims = ax2.get_xlim()
        ylims = ax2.get_ylim()
        
        # fit a normal distribution to the data
        norm_mu, norm_std = stats.norm.fit(_obj.residuals.dropna())
        pdf_x = np.linspace(xlims[0], xlims[1], 100)
        bluebelt.graph.defaults.dotted_line(pdf_x, stats.norm.pdf(pdf_x, norm_mu, norm_std), ax=ax2)
        # ax2.plot(pdf_x, stats.norm.pdf(pdf_x, norm_mu, norm_std), **style.patterns.normal_plot)

        # histogram x label
        ax2.set_xlabel('residuals distribution')
                
        ax2.set_ylim(ylims[0], ylims[1]*1.5)
        ax2.spines['left'].set_visible(False)
        ax2.spines['right'].set_visible(False)

        ax2.text(0.02, 0.7, f'D’Agostino-Pearson\nstatistic: {_obj.statistic:1.2f}\np: {_obj.p_value:1.2f}', transform=ax2.transAxes, **style.patterns.statistics)
    else:
        gridspec = fig.add_gridspec(nrows=1, ncols=1)
        
    ax1 = fig.add_subplot(gridspec[0, 0], zorder=50)
    
    # set indices
    _index = bluebelt.core.index.IndexToolkit(_obj.series.index).alt()
    _observations_index = bluebelt.core.index.IndexToolkit(_obj.within_bounds.index).alt()
    _out_of_bounds_index = bluebelt.core.index.IndexToolkit(_obj.out_of_bounds.index).alt()
    _within_bounds_index = bluebelt.core.index.IndexToolkit(_obj.within_bounds.index).alt()
    _outliers_index = bluebelt.core.index.IndexToolkit(_obj.outliers.index).alt()
    
    
    # pattern
    bluebelt.graph.defaults.line(_index, _obj.result.values, ax=ax1, zorder=95, label='pattern')
    
    if bounds:
        # observations
        bluebelt.graph.defaults.observations(_observations_index, _obj.within_bounds.values, ax=ax1, zorder=99, label='observations')

        # bounds
        bluebelt.graph.defaults.fill_between(_index, _obj.lower.values, _obj.upper.values, ax=ax1, zorder=90, label=f'{(_obj.confidence * 100):1.0f}% bounds')
        bluebelt.graph.defaults.dotted_line(_index, _obj.lower.values, ax=ax1, zorder=92, label=None)
        bluebelt.graph.defaults.dotted_line(_index, _obj.upper.values, ax=ax1, zorder=92, label=None)
        
        # out of bounds
        bluebelt.graph.defaults.out_of_bounds(_out_of_bounds_index, _obj.out_of_bounds.values, ax=ax1, zorder=85, label='out of bounds')
    else:
        # observations
        bluebelt.graph.defaults.observations(_within_bounds_index, _obj.within_bounds.values, ax=ax1, zorder=99, label='observations')

        # out of bounds
        bluebelt.graph.defaults.observations(_within_bounds_index, _obj.out_of_bounds.values, ax=ax1, zorder=85, label='out of bounds')

    # outliers
    bluebelt.graph.defaults.outliers(_outliers_index, _obj.outliers.values, ax=ax1, zorder=80, label='outliers')
        
    # labels
    ax1.set_title(f'{_obj.result.name}', **style.patterns.title)
    ax1.set_ylabel('value')

    # set xticks
    bluebelt.helpers.xticks.set_xticks(ax=ax1, index=_obj.series.index, location=_index, group=group)

    ax1.set_xlim(xlim)
    ax1.set_ylim(ylim)

    # legend
    if legend:
        ax1.legend(loc='best')
    
    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig